{
  "type": "record",
  "Aspect": {
    "name": "lifecycleStageTypeKey",
    "keyForEntity": "lifecycleStageType",
    "entityCategory": "core",
    "entityAspects": [
      "lifecycleStageTypeInfo",
      "status"
    ],
    "entityDoc": "Defines a lifecycle stage that entities can be placed in (e.g., Proposed, Certified, Archived). Controls search visibility and transition policies."
  },
  "name": "LifecycleStageTypeKey",
  "namespace": "com.linkedin.pegasus2avro.metadata.key",
  "fields": [
    {
      "type": "string",
      "name": "id",
      "doc": "Unique identifier for the lifecycle stage type, e.g. \"PROPOSED\", \"CERTIFIED\", \"ARCHIVED\"."
    }
  ],
  "doc": "Key for a Lifecycle Stage Type"
}